// ..................................................................
//
//                          mdMailSlot
//
//               Copyright 1997 by Martin Djerns
//
// ..................................................................
// Initial Date : 2 April 1997 - MD
// Version 1.0  : 3 April 1997 - MD
// ..................................................................

unit mdMailSlot;

interface

{$LONGSTRINGS ON}  // Equal {$H+}

uses
  ExtCtrls, Windows, SysUtils, Classes;

Const
  // Missing constant!
  MAILSLOT_WAIT_FOREVER = $FFFFFFFF;

  // Error strings
  sCantOpen             = 'Cannot open mailslot';
  sNotActive            = 'Not active';
  sGetMailSlotInfoError = 'Error in GetMailSlotInfo';
  sNoMessages           = 'No messages waiting';
  sReadError            = 'Error reading message';
  sExists               = 'Mailslot allready exists';
  sBadName              = 'Bad mailslot name';

  // Default values
  msDefActive = False;
  msDefServer  = '.';
  msDefSlot = 'mdMailSlot1';
  msDefMaxSize = 0;

type
// .........................................................
//                Component types
// .........................................................

  TmsMessageAvail = Procedure (Sender : TObject; Msg : String) of Object;

// .........................................................
//       EXCEPTION :     EmdMailSlot
// .........................................................

  EmdMailSlot = Class(Exception);

// .........................................................
//                       TmdCustomMailSlot
// .........................................................

  TmdCustomMailSlot = class(TComponent)
  private          { Private declarations }

    // Property storage variables
    FActive : Boolean;
    FHandle : THandle;
    FMaxSize : DWord;
    FServer : String;
    FSlot : String;

    FWaiting : DWord;
    FNextSize : DWord;

    // Event storage variables
    FOpen : TNotifyEvent;
    FClose : TNotifyEvent;

    // Misc values
    FSlotChanged : Boolean;

    // Property modification functions
    Procedure SetActive(Value : Boolean);
    Procedure SetSlot(Const Value : String);
    Procedure SetServer(Const Value : String);
    Procedure SetMaxSize(Value : DWord);

    Function GetWaiting : DWord;
    Function GetNextSize : DWord;
  protected        { Protected declarations }

    // Event process functions
    Procedure DoOpen; Virtual;
    Procedure DoClose; Virtual;

    // Special functions
    Procedure PerformError(Const Err : String); Virtual;
    Procedure UpdateSlotInfo;
  public  { Public declarations }
    // Default functions
    Constructor Create(AOwner : TComponent); Override;
    Procedure SetName(const NewName: TComponentName); Override;

    // Special functions
    Procedure Open; Virtual;
    Procedure Close; Virtual;
    Function ReadNext : String; Virtual;

    // properties - Read Only
    Property Handle : THandle Read FHandle;
    Property Waiting : DWord Read GetWaiting;
    Property NextSize : DWord Read GetNextSize;
    // properties - Read/Write
    Property Active : Boolean Read FActive Write SetActive Default msDefActive;
    Property Server : String Read FServer Write SetServer;
    Property Slot : String Read FSlot Write SetSlot Stored FSlotChanged;
    Property MaxSize : DWord Read FMaxSize Write SetMaxSize Default msDefMaxSize;

    // Events
    Property OnOpen : TNotifyEvent Read FOpen Write FOpen;
    Property OnClose : TNotifyEvent Read FClose Write FClose;
  published        { Published declarations }
  end;

// .........................................................
//                       TmdMailSlot
// .........................................................

  TmdMailSlot = class(TmdCustomMailSlot)
  Published
    // Properties
    Property Active;
    Property Server;
    Property Slot;
    Property MaxSize;
    // Events
    Property OnOpen;
    Property OnClose;
  end;

// .........................................................
//                       TmdAutoMailSlot
// .........................................................
// Note : This class are created for automatically getting
//        a Delphi event when something comes into the
//        mailslot. While I not yet know about any way of
//        getting this information from Windows, have I
//        implemented a class which are using a TTimer
//        class for checking every second if anything is
//        arrived.
//        In case of that I find a bether way (registere a
//        mailslot change i Windows o.e.) I'll keep the
//        functions (interface) in this class as thay are,
//        and implement the function directly in
//        TmdCustomMailSlot. This class will then get the
//        functionallity from the base class, but it will
//        be backward compatible.
// .........................................................

  TmdAutoMailSlot = Class(TmdMailSlot)
  Private
    FTimer : TTimer;
    FMessageList : TStringList;
    FMessageAvail : TmsMessageAvail;
    Procedure TimerProc(Sender : TObject);
  Protected
    Procedure AddMessage(Const Msg : String);
    Function GetMessage : String;
    Procedure DoMessageAvail(Const Msg : String); Virtual;
  Public
    Constructor Create(AOwner : TComponent); Override;
    Procedure Open; Override;
    Procedure Close; Override;
  Published
    Property OnMessageAvail : TmsMessageAvail Read FMessageAvail Write FMessageAvail;
  end;

// .........................................................
//                       Functions
// .........................................................

// Send a single message to a mailslot
// If Server is "." then the current machine are used
// If Server is "*" then the current domain are used
Function SendToMailSlot(Const Server, Slot, Mail : String) : Boolean;

procedure Register;

implementation

// .........................................................
//                      SendToMailSlot
// .........................................................

Function SendToMailSlot(Const Server, Slot, Mail : String) : Boolean;
Var
  hToSlot : THandle;
  BytesWritten : DWord;
  FullSlot : String;
Begin
  Result := False; // Default value

  FullSlot := '\\'+Server+'\mailslot\'+Slot;  // MailSlot string
  hToSlot := CreateFile(
    PChar(FullSlot),
    GENERIC_WRITE,
    FILE_SHARE_READ,
    NIL,
    OPEN_EXISTING,
    FILE_ATTRIBUTE_NORMAL,
    -1                    );
  If hToSlot = INVALID_HANDLE_VALUE Then
    Exit; // Error
  Try
    BytesWritten := 0;
    If (NOT WriteFile(hToSlot,
                      Pointer(Mail)^,
                      Length(Mail),
                      BytesWritten,
                      NIL))         OR
        (BytesWritten <> Length(Mail)) Then
      Exit; // Error

    Result := True;  // Everthing went good
  Finally
    CloseHandle(hToSlot);
  end;
end;

// .........................................................
//                    TmdCustomMailSlot
// .........................................................

// Default functions

Constructor TmdCustomMailSlot.Create(AOwner : TComponent);
Begin
  Inherited Create(AOwner);
  FSlotChanged := False;
  FHandle := INVALID_HANDLE_VALUE;
  FSlot := msDefSlot;
  FServer := msDefServer;
  FMaxSize := msDefMaxSize;

  If msDefActive Then
    Open
  else
    Close;
end;

Procedure TmdCustomMailSlot.SetName(const NewName: TComponentName);
Begin
  Inherited SetName(NewName);
  If NOT FSlotChanged Then
    FSlot := NewName;
end;

// Special functions

Procedure TmdCustomMailSlot.PerformError(Const Err : String);
Begin
  Raise EmdMailSlot.Create(Err);
end;

Procedure TmdCustomMailSlot.Open;
Var
  MailSlotName : String;
  Buffer : String;
  ReadSize : DWord;
Begin
  If Active Then
    Close;
  MailSlotName := '\\'+FServer+'\mailslot\'+FSlot;
  If NOT (csDesigning IN ComponentState) Then
  Begin
    FHandle := CreateMailSlot(PChar(MailSlotName),
                            MaxSize,
                            MAILSLOT_WAIT_FOREVER,
                            NIL);
    FActive := Handle <> INVALID_HANDLE_VALUE;
    If NOT Active Then
    Begin
      Case GetLastError of
        ERROR_ALREADY_EXISTS : PerformError(sExists);
        ERROR_BAD_PATHNAME   : PerformError(sBadName);
      else
        PerformError(sCantOpen);
      end;
    end
    else
      DoOpen;
  end
  else
    FActive := True;
end;

Procedure TmdCustomMailSlot.Close;
Begin
  If Active Then
  Begin
    CloseHandle(FHandle);
    FHandle := INVALID_HANDLE_VALUE;
    FActive := False;
    DoClose;
  end;
end;

Procedure TmdCustomMailSlot.UpdateSlotInfo;
Begin
  If NOT Active Then
    PerformError(sNotActive);

 If NOT GetMailSlotInfo(Handle,NIL, FNextSize, @FWaiting, NIL) Then
   PerformError(sGetMailSlotInfoError);
end;

Function TmdCustomMailSlot.ReadNext : String;
Var
  ReadSize : DWord;
Begin
  UpdateSlotInfo;
  If FWaiting = 0 Then
    PerformError(sNoMessages);
  SetLength(Result,FNextSize);
  ReadFile(Handle,PChar(Result)^,FNextSize,ReadSize,NIL);
  If FNextSize <> ReadSize Then
    PerformError(sReadError);
end;

// Property modification functions

Procedure TmdCustomMailSlot.SetActive(Value : Boolean);
Begin
  If Value Then
    Open
  else
    Close;
end;

Procedure TmdCustomMailSlot.SetServer(Const Value : String);
Begin
  If Value = FServer Then
    Exit;
  FServer := Value;
  If Active Then
    Open;
end;

Procedure TmdCustomMailSlot.SetSlot(Const Value : String);
Begin
  If Value = FSlot Then
    Exit;
  FSlotChanged := True;
  FSlot := Value;
  If Active Then
    Open;
end;

Procedure TmdCustomMailSlot.SetMaxSize(Value : DWord);
Begin
  If Value = FMaxSize Then
    Exit;
  FMaxSize := Value;
  If Active  Then
    Open;
end;

Function TmdCustomMailSlot.GetWaiting : DWord;
Begin
  Result := 0;
  // I want that it is possible to check on
  // waiting without testing Active first
  If NOT Active Then
    Exit;
  UpdateSlotInfo;
  Result := FWaiting;
end;

Function TmdCustomMailSlot.GetNextSize : DWord;
Begin
  UpdateSlotInfo;
  Result := FNextSize;
end;

// Event process functions

Procedure TmdCustomMailSlot.DoOpen;
Begin
  If Assigned(FOpen) Then
    FOpen(Self);
end;

Procedure TmdCustomMailSlot.DoClose;
Begin
  If Assigned(FClose) Then
    FClose(Self);
end;

// .........................................................
//                     TmdAutoMailSlot
// .........................................................

Constructor TmdAutoMailSlot.Create(AOwner : TComponent);
Begin
  FTimer := NIL;
  FMessageList := NIL;
  Inherited Create(AOwner);
end;

Procedure TmdAutoMailSlot.AddMessage(Const Msg : String);
Begin
  If FMessageList = NIL Then
    FMessageList := TStringList.Create;
  FMessageList.Add(Msg);
end;

Function TmdAutoMailSlot.GetMessage : String;
Begin
  Result := '';
  If FMessageList <> NIL Then
  Begin
    If FMessageList.Count > 0 Then
    Begin
      Result := FMessageList[0];
      FMessageList.Delete(0);
    end;
    If FMessageList.Count = 0 Then
    Begin
      FMessageList.Free;
      FMessageList := NIL;
    end;
  end;
end;

Procedure TmdAutoMailSlot.Open;
Begin
  Inherited Open;
  If Active AND (NOT (csDesigning IN ComponentState))Then
  Begin
    If FTimer = NIL Then
      FTimer := TTimer.Create(Self);
    FTimer.Interval := 1000;
    FTimer.OnTimer := TimerProc;
  end;
end;

Procedure TmdAutoMailSlot.Close;
Begin
  If FTimer <> NIL Then
    FTimer.Free;
  FTimer := NIL;
  Inherited Close;
end;

Procedure TmdAutoMailSlot.TimerProc(Sender : TObject);
Begin
  While Waiting > 0 do
    AddMessage(ReadNext);

  If (FMessageList <> NIL) AND (FMessageList.Count > 0) Then
    DoMessageAvail(GetMessage);
end;

Procedure TmdAutoMailSlot.DoMessageAvail(Const Msg : String);
Begin
  If Assigned(FMessageAvail) Then
    FMessageAvail(Self, Msg);
end;

// .........................................................
//                        Register
// .........................................................

procedure Register;
begin
  RegisterComponents('mdVCL', [TmdMailSlot, TmdAutoMailSlot]);
end;

end.
